#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <memory.h>

#define MATRIX_SIZE 5

int active = 0;

void child_handler(int signal, siginfo_t * data, void * unused ) {
    int retVal;

    if( signal == SIGCHLD ) {
        switch( data->si_code ) {
        case CLD_DUMPED:
            /* Oops... A crash. */
            printf( "Child crashed :(\n" );
        case CLD_EXITED:
            /* Child exited normally. */
        case CLD_KILLED:
            /* Child was killed. */
            
            /* Clean up... */
            waitpid( data->si_pid, &retVal, 0 );
            active--;
            /* Now, retVal contains the exit value of the child. */


            break;
        default:
            /* Something happened we don't care about. */
            break;
        }
    }
}

void init_matrix( int matrix[], int offset ) {
    int x, y;
    for( y = 0; y < MATRIX_SIZE; y++ ) {
        for( x = 0; x < MATRIX_SIZE; x++ ) {
            matrix[y * MATRIX_SIZE + x] = y + x + offset;
        }
    }
}

int calc_elem( int matrixA[], int matrixB[], int x, int y ) {
    int j;
    int retVal = 0;

    for( j = 0; j < MATRIX_SIZE; j++ ) {
        retVal += matrixA[y * MATRIX_SIZE + j] * matrixB[j * MATRIX_SIZE + x];
    }
    return retVal;
}

void calc_row( int matrixC[], int matrixA[], int matrixB[], int row ) {
    int i;
    for( i = 0; i < MATRIX_SIZE; i++ ) {
        matrixC[row * MATRIX_SIZE + i] = calc_elem( matrixA, matrixB, i, row);
    }
}

void print_matrix( int matrix[] ) {
    int x, y;
    for( y = 0; y < MATRIX_SIZE; y++ ) {
        printf( " [" );
        for( x = 0; x < MATRIX_SIZE; x++ ) {
            printf( "%4d ", matrix[y * MATRIX_SIZE + x] );
        }
        printf( "]\n" );
    }
}

int main() {
    struct sigaction chld_handler;
    int pid, x, y;
    int matrixA[MATRIX_SIZE * MATRIX_SIZE];
    int matrixB[MATRIX_SIZE * MATRIX_SIZE];
    int matrixC[MATRIX_SIZE * MATRIX_SIZE];

    /* Initialize the matricies. */
    init_matrix( matrixA, 0 );
    init_matrix( matrixB, 1 );
    memset( matrixC, 0x00, sizeof(matrixC) );

    /* Set up the signal handler to catch the children as they exit. */
    memset( &chld_handler, 0x00, sizeof(struct sigaction) );
    chld_handler.sa_sigaction = child_handler;
    chld_handler.sa_flags = SA_SIGINFO;
    sigaction( SIGCHLD, &chld_handler, NULL );

    printf( "Matrix A:\n" );
    print_matrix( matrixA );
    printf( "Matrix B:\n" );
    print_matrix( matrixB );

    /* Start those children running... */
    for( y = 0; y < MATRIX_SIZE; y++ ) {
        for( x = 0; x < MATRIX_SIZE; x++ ) {
            pid = fork();
            if( pid == 0 ) {
                /* We are the child. */
                int value = calc_elem( matrixA, matrixB, x, y );
                return value;
            } else {
                /* We are the parent. */
                active++;
            }
        }
    }

    /* Now, we wait for all the children to finish. */
    while( active );

    printf( "Result:\n" );
    print_matrix( matrixC );
    return 0;
}
